#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include "opdf.h"

/*************************************************************************
  Basic operations: arrays etc.
 *************************************************************************/

static int     Oerrno;

typedef struct 
{
  int *size, *length;
  void **data;
  int eltsize;
} array_t;

#define ARRAY(x) \
   static array_t A##x = { &s##x, &n##x, (void**) &x, sizeof(x[0]) }

static void Agrow(array_t *A)
{
  if (A->length[0]==A->size[0])
  {
    A->size[0] += A->size[0]/2 + 10;
    A->data[0]= realloc(A->data[0], A->eltsize*A->size[0]);
  }
}

static void Ainit(array_t *A)
{
  if (A->size[0]==0)
  {
    A->size[0]= 50;
    A->data[0]= malloc(A->size[0]*A->eltsize);
  }
  A->length[0]= 0;
}

#define unt32(A,B,C,D)  (((A)<<24) | ((B)<<16) | ((C)<<8) | (D))


/**************************************************************************
  Input System:
    Provides two methods of input, one from a file and one from the
    memory. The buffer provided for the memory stream is managed by
    the user.
 **************************************************************************/

static int                   Itype;
static char                 *Ifn;
static FILE                 *Ifile;
static off_t                 Ioffset;
static const unsigned char  *Idata;
static size_t                Isize;


static int Iclose(int ret)
{
  if (ret) Oerrno= ret;
  if (Ifile) fclose(Ifile);
  Ifile= NULL;
  if (Ifn) free(Ifn);
  Ifn= NULL;
  return ret;
}

static int Iseek(off_t offset)
{
  if (Itype=='F')
  {
    if (fseek(Ifile, offset, SEEK_SET))
      return 1;
  }
  else
  {
    if (offset>Isize) return 1;
  }
  Ioffset= offset;
  return 0;
}

static int Iskip(size_t sz)
{
  if (Itype=='F')
  {
    if (fseek(Ifile, sz, SEEK_CUR))
      return 1;
  }
  else
  {
    if (Ioffset+sz>Isize) return 1;
  }
  Ioffset+= sz;
  return 0;
}

static int Ibytes(size_t sz, unsigned char *data)
{
  if (Itype=='F') 
  {
    int R;
    R= fread(data, 1, sz, Ifile);
    if (R!=sz) return 1;
  }
  else
  {
    if (Ioffset+sz>Isize) return 1;
    memcpy(data, Idata+Ioffset, sz);
  }
  Ioffset+= sz;
  return 0;
}

static unsigned char Iunt8()
{
  unsigned char b[1];
  if (Ibytes(1,b)) return 0xFF;
  return b[0];
}

static unsigned int Iunt32()
{
  unsigned char b[4];
  if (Ibytes(4,b)) return 0xFFffFFff;
  return unt32(b[0],b[1],b[2],b[3]);
}

/************************************************************************
  Output system:
   Only file output is necessary for my purposes. I keep track of the
   file size as I write things to it because I record offsets for the
   xref table.
 ************************************************************************/

static FILE   *Ofile;
static size_t  Osize;

static int Obytes(const unsigned char *data, size_t len)
{
  int R;
  if (!Ofile) return 1;
  while(len)
  {
    R= fwrite(data, 1, len, Ofile);
    if (R<=0) return 1;
    data+= R;
    len-= R;
    Osize+= R;
  }
  return 0;
}

static int Ostr(const char *str)
{
  return Obytes((const unsigned char*) str, strlen(str));
}

static int Oinit(const char *fn)
{
  Oerrno= 0;
  Ofile= fopen(fn,"wb");
  if (!Ofile) { Oerrno= errno; return 1; }
  Osize= 0;
  return 0;
}

static int OcopyF(size_t sz)
{
  unsigned char buf[1024];
  int R, cp;
  while(sz)
  {
    cp= sizeof(buf);
    if (cp>sz) cp= sz;
    R= fread(buf, 1, cp, Ifile);
    if (R<=0) return 1;
    if (Obytes(buf,R)) return 1;
    sz -= R;
  }
  return 0;
}

static int Ocopy(size_t sz)
{
  int R;
  if (Itype=='F')
    return OcopyF(sz);
  if (Ioffset+sz>Isize) return 1;
  R= Obytes(Idata+Ioffset, sz);
  Ioffset+= sz;
  return R;
}

static int Oprint(const char *fmt, ...)
{
  va_list ap;
  int R;
  va_start(ap,fmt);
  R= vfprintf(Ofile, fmt, ap);
  va_end(ap);
  if (R<0)
    return 1;
  Osize+= R;
  return 0;
}

/**********************************************************************
 Buffer objects:
   These are used for buffering things such as resource declarations
   or page contents. 
 **********************************************************************/
typedef struct
{
  char **str;
  int nstr,sstr;
  array_t array;
} buffer_t;

#define BUFFER(x)  static buffer_t B##x= { NULL, 0, 0, \
    { &B##x.sstr, &B##x.nstr, (void**) &B##x.str, sizeof(char*) } }

static void Binit(buffer_t *B) { Ainit(&B->array); }

static void Bprintv(buffer_t *B, const char *fmt, va_list ap)
{
  int R;
  static char buf[2000];
  R= vsnprintf(buf,sizeof(buf)-1,fmt, ap);
  buf[R]= 0;
  Agrow(&B->array);
  B->str[B->nstr++]= strdup(buf);
}

static int Bsize(buffer_t *B)
{
  int S,i;
  for(S=0,i=0;i<B->nstr;i++)
   S+= strlen(B->str[i]);
  return S;
}

static void Obuffer_d(buffer_t *B)
{
  int i;
  for(i=0;i<B->nstr;i++)
  {
    Ostr(B->str[i]);
    free(B->str[i]);
  }
  B->nstr= 0;
}

/*************************************************************************
 Cross reference table:
   This table holds the offsets of each object in the PDF file. 
 *************************************************************************/

static int             nxref, sxref,*xref;
                 ARRAY(xref);
                  // This is set when the xref table is output.
                  // The file trailer refers to this offset.
static size_t          xref_offset;

static void init_xref()
{
  Ainit(&Axref);
  nxref= 1;
  xref[0]= 0;
}

static int reserve_object()
{
  Agrow(&Axref);
  nxref++;
  return nxref-1;
}

static void begin_object(int obno)
{
  xref[obno]= Osize;
  Oprint("%d 0 obj\n", obno);
}

static void end_object()   { Ostr("endobj\n"); }
static void begin_stream() { Ostr("stream\n"); }
static void end_stream()   { Ostr("\nendstream\n"); }

static void output_xref()
{
  int i;
  xref_offset= Osize;
  Oprint("xref\n"
         "0 %d\n"
         "0000000000 65535 f\r\n", nxref);
  for(i=0;i<nxref;i++)
    Oprint("%010d 00000 n\r\n", xref[i]);
}

/****************************************************************************
 Pages:
   The list of pages are stored in a root page-node. This node contains
   the object numbers for all pages. The catalog object then refers to
   this object.
 ****************************************************************************/

                 // root node for the page tree
static int             Jpages_root;
                 // the Catalog object, which refers to Jpages_root 
                 // the file trailer points to this object.
static int             Jcatalog;
static int             npages, spages, *pages;
                 ARRAY(pages);

static void init_pages()
{
  Jpages_root= reserve_object();
  Ainit(&Apages);
  Jcatalog= reserve_object();
  begin_object(Jcatalog);
  Oprint(" << /Type Catalog\n"
         "    /Pages %d 0 R\n"
         " >>\n", Jpages_root);
  end_object();
}

static void record_page(int pageob)
{
  Agrow(&Apages);
  pages[npages++]= pageob;
}

static void output_pages_root()
{
  int i;
  begin_object(Jpages_root);
  Oprint(" << /Type Pages\n"
         "    /Kids [ ");
  for(i=0;i<npages;i++)
    Oprint(" %d 0 R\n", pages[i]);
  Oprint(" ] \n"
         " /Count %d\n"
         " >>\n", npages);
  end_object();
}

static void output_trailer()
{
  Oprint("trailer\n"
         " << /Size %d\n"
         "    /Root %d 0 R\n"
         " >>\n"
         "startxref\n"
         "%d\n"
         "%c%cEOF\n",
            nxref, Jcatalog, xref_offset, '%', '%');
}

/**************************************************************************
  A resource dictionary defines objects to be used in drawing commands.
  For example, an image object with a given object number is assigned
  a name here so that the drawing command can access it. It works like
  this:
    <<  /Xobject << /Im1  10 0 R  /Im2 /11 0 R >> >>
  There are other things such as fonts etc, but I won't do those yet.

  This dictionary is normally a part of the page dictionary under the
  name of 'Resources'. However, it can be output as a seperate object
  and then referred to using the '/Resources 12 0 R' syntax etc. This
  is what I do.
 **************************************************************************/

typedef struct
{
  int object;
  int type;
} resource_t;
                   // This object holds the resources dictionary
                   // for the current page.
static int             Jresources;
static resource_t     *resources;
static int             nresources, sresources;
                 ARRAY(resources);

static void init_resources()
{
  Ainit(&Aresources);
}

static int get_resource(int object,int type)
{
  int i;
  for(i=0;i<nresources;i++)
   if (resources[i].object==object) return object;
  Agrow(&Aresources);
  resources[nresources].object= object;
  resources[nresources].type= type;
  nresources++;
  return object;
}

static void output_resources()
{
  int i;
  begin_object(Jresources);
  Oprint(" << /XObject <<\n");
  for(i=0;i<nresources;i++)
   Oprint("  /RE%d  %d 0 R\n",
                resources[i].object, 
                resources[i].object);
  Oprint(" >>\n >>\n");
  end_object(Jresources);
}

/*************************************************************************
 Contents are stored in stream objects. These are simple strings which
 describe the operations to be done.
 *************************************************************************/

static int             Jcontents;
                BUFFER(contents);

static void init_contents()
{
  Binit(&Bcontents);
}

static void Content(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  Bprintv(&Bcontents, fmt, ap);
  va_end(ap);
}

static void output_contents()
{
  begin_object(Jcontents);
  Oprint(" << /Length %d  >> \n", Bsize(&Bcontents));
  begin_stream();
  Obuffer_d(&Bcontents);
  end_stream();
  end_object();
}

/***************************************************************************
  Page Object:
    This holds the resources and the contents.
 ***************************************************************************/

void opdf_begin_page(int width, int height)
{
  int Jpage;
  Jpage= reserve_object();
  Jresources= reserve_object();
  Jcontents= reserve_object();

  begin_object(Jpage);
  Oprint(" << /Type Page\n"
         "  /Parent %d 0 R\n"
         "  /MediaBox [ 0 0 %d %d ] \n"
         "  /Resources %d 0 R\n"
         "  /Contents  %d 0 R\n"
         "  >>\n",  Jpages_root, width, height,
                     Jresources, Jcontents);
  end_object();
  record_page(Jpage);
  init_resources();
  init_contents();
}

void opdf_end_page()
{
  output_resources();
  output_contents();
}

/*************************************************************************
 Top level file structure code.
 *************************************************************************/

int opdf_begin(char *fn)
{
  if (Oinit(fn)) return 1;
  init_xref();
  static const unsigned char hdr[]=
  {
    '%', 'P', 'D', 'F', '-', '1', '.', '5', '\n',
    '%', 0xb5, 0xed, 0xae, 0xfb, '\n', 0
  };
  Ostr((const char*) hdr);

  init_pages();
  
  return 0;
}

int opdf_end()
{
  output_pages_root();
  output_xref();
  output_trailer();
  if (Ofile) fclose(Ofile);
  Ofile= NULL;
  return 0;
}


/**************************************************************************
 PNG image loading code.
 **************************************************************************/

#define png_IHDR   unt32('I','H','D','R')
#define png_IDAT   unt32('I','D','A','T')

#define png_SIG1   unt32(137, 'P', 'N', 'G')
#define png_SIG2   unt32('\r', '\n', 26, '\n')


static int read_png_info(int *RJimage, int *Rwidth, int *Rheight)
{
  int width, height, Jimage;
  errno= EBADF;
  if (Iunt32()!=png_SIG1 || Iunt32()!=png_SIG2)
    return EBADF;  
  
  if (Iunt32()!=13) return EBADF; // bad length in headr..
  if (Iunt32()!=png_IHDR) return EBADF;
  width= Iunt32();
  height= Iunt32();
  if (Iunt8()!=8) return EBADF; // bit depth
  if (Iunt8()!=2) return EBADF; // color type
  if (Iunt8()!=0) return EBADF; // compression method
  if (Iunt8()!=0) return EBADF; // filter method
  if (Iunt8()!=0) return EBADF; // interlace method
  Iskip(4); // skip over the CRC


  int Jimg, Jlen;
  Jimg= reserve_object();
  Jlen= reserve_object();
  begin_object(Jimg);
  Oprint(" << /Filter /FlateDecode\n"
   "/Type /XObject\n"
   "/Subtype /Image\n"
   "/Width %d\n"
   "/Height %d\n"
   "/ColorSpace /DeviceRGB\n"
   "/Interpolate true\n"
   "/BitsPerComponent 8\n"
   "/Length %d 0 R\n"
   "/DecodeParms << /Predictor 15\n"
   "    /Colors 3\n"
   "    /Columns %d\n"
   "    /BitsPerComponent 8\n"
   "  >>\n>>\n",
      width, height, Jlen, width);
  begin_stream();

  unsigned int total_length, cL, cT;
  total_length= 0;

  while(1)
  {
    cL= Iunt32();
    cT= Iunt32();
    if (cL==-1 || cT==-1) break;
    if (cT==png_IDAT)
    {
      total_length+= cL;
      Ocopy(cL);
    }
    else
    {
      Iskip(cL);
    }
    Iskip(4);
  }
  end_stream();
  end_object();
  begin_object(Jlen);
  Oprint("  %d\n", total_length);
  end_object();
  if (total_length==0) return EBADF;
  if (RJimage) *RJimage= Jimg;
  if (Rwidth) *Rwidth= width;
  if (Rheight) *Rheight= height;
  return 0;
}


int opdf_pngfile(char *fn,int *Rimg, int *Rw, int *Rh)
{
  int R;
  Itype= 'F';
  Ifn= strdup(fn);
  Ifile= fopen(fn, "rb");
  if (!Ifile) return Iclose(errno);
  Ioffset= 0;
  R= read_png_info(Rimg, Rw, Rh);
  if (R) return Iclose(R);
  return Iclose(0);
}

int opdf_pngmem
   (const unsigned char *data, size_t len, int *Rimg, int *Rw, int *Rh)
{
  int R;
  Itype= 'M';
  Ioffset= 0;
  Isize= len;
  Idata= data;
  R= read_png_info(Rimg, Rw, Rh);
  if (R) return R;
  return 0;
}


/************************************************************************
 Drawing Functions
 ************************************************************************/

static void Ctranslate(double x,double y)
{
  Content(" 1 0 0 1 %f %f cm ",x,y);
}

static void Cscale(double w, double h)
{
  Content(" %f 0 0 %f 0 0 cm ", w, h);
}

void opdf_draw_image(int img, int x, int y, int w, int h)
{
  img= get_resource(img, 'I');
  Content(" q ");
  Ctranslate(x,y);
  Cscale(w,h);
  Content(" /RE%d Do Q\n",img);
}
